@@ -1,87 +1,88 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
+import hashlib
+
 from sql import Null
 
+from trytond.i18n import gettext
 from trytond.model import Workflow, Model, ModelView, ModelSQL, fields, Check
+from trytond.model.exceptions import AccessError
 from trytond.pyson import Eval, Bool, If
 from trytond.transaction import Transaction
 from trytond.pool import Pool
 from trytond.wizard import Wizard, StateView, StateTransition, Button
 
-__all__ = ['Inventory', 'InventoryLine',
-    'Count', 'CountSearch', 'CountQuantity']
-
-STATES = {
-    'readonly': Eval('state') != 'draft',
-}
-DEPENDS = ['state']
-INVENTORY_STATES = [
-    ('draft', 'Draft'),
-    ('done', 'Done'),
-    ('cancel', 'Canceled'),
-    ]
+from .exceptions import (
+    InventoryValidationError, InventoryCountWarning, InventoryFutureWarning)
 
 
 class Inventory(Workflow, ModelSQL, ModelView):
     'Stock Inventory'
     __name__ = 'stock.inventory'
     _rec_name = 'number'
-    number = fields.Char('Number', readonly=True)
+
+    _states = {
+        'readonly': Eval('state') != 'draft',
+        }
+    _depends = ['state']
+
+    number = fields.Char('Number', readonly=True,
+        help="The main identifier for the inventory.")
     location = fields.Many2One(
         'stock.location', 'Location', required=True,
         domain=[('type', '=', 'storage')], states={
             'readonly': (Eval('state') != 'draft') | Eval('lines', [0]),
             },
-        depends=['state'])
+        depends=['state'],
+        help="The location inventoried.")
     date = fields.Date('Date', required=True, states={
             'readonly': (Eval('state') != 'draft') | Eval('lines', [0]),
             },
-        depends=['state'])
-    lost_found = fields.Many2One(
-        'stock.location', 'Lost and Found', required=True,
-        domain=[('type', '=', 'lost_found')], states=STATES, depends=DEPENDS)
+        depends=['state'],
+        help="The date of the stock count.")
     lines = fields.One2Many(
         'stock.inventory.line', 'inventory', 'Lines',
         states={
-            'readonly': (STATES['readonly'] | ~Eval('location')
+            'readonly': (_states['readonly'] | ~Eval('location')
                 | ~Eval('date')),
             },
-        depends=['location', 'date'] + DEPENDS)
+        depends=['location', 'date'] + _depends)
     empty_quantity = fields.Selection([
             (None, ""),
             ('keep', "Keep"),
             ('empty', "Empty"),
-            ], "Empty Quantity", states=STATES, depends=DEPENDS,
-        help="How lines without quantity are handled.")
+            ], "Empty Quantity", states=_states, depends=_depends,
+        help="How lines without a quantity are handled.")
     company = fields.Many2One('company.company', 'Company', required=True,
         states={
             'readonly': (Eval('state') != 'draft') | Eval('lines', [0]),
             },
-        depends=['state'])
-    state = fields.Selection(
-        INVENTORY_STATES, 'State', readonly=True, select=True)
+        depends=['state'],
+        help="The company the inventory is associated with.")
+    state = fields.Selection([
+            ('draft', "Draft"),
+            ('done', "Done"),
+            ('cancelled', "Cancelled"),
+            ], "State", readonly=True, select=True,
+        help="The current state of the inventory.")
+
+    del _states, _depends
 
     @classmethod
     def __setup__(cls):
         super(Inventory, cls).__setup__()
         cls._order.insert(0, ('date', 'DESC'))
-        cls._error_messages.update({
-                'delete_cancel': ('Inventory "%s" must be canceled before '
-                    'deletion.'),
-                'unique_line': ('Line "%s" is not unique '
-                    'on Inventory "%s".'),
-                })
         cls._transitions |= set((
                 ('draft', 'done'),
-                ('draft', 'cancel'),
+                ('draft', 'cancelled'),
                 ))
         cls._buttons.update({
                 'confirm': {
-                    'invisible': Eval('state').in_(['done', 'cancel']),
+                    'invisible': Eval('state').in_(['done', 'cancelled']),
                     'depends': ['state'],
                     },
                 'cancel': {
-                    'invisible': Eval('state').in_(['cancel', 'done']),
+                    'invisible': Eval('state').in_(['cancelled', 'done']),
                     'depends': ['state'],
                     },
                 'complete_lines': {
@@ -98,11 +99,21 @@ class Inventory(Workflow, ModelSQL, ModelView):
     def __register__(cls, module_name):
         super(Inventory, cls).__register__(module_name)
 
+        cursor = Transaction().connection.cursor()
         table = cls.__table_handler__(module_name)
+        sql_table = cls.__table__()
 
         # Add index on create_date
         table.index_action('create_date', action='add')
 
+        # Migration from 5.4: remove lost_found
+        table.not_null_action('lost_found', 'remove')
+
+        # Migration from 5.6: rename state cancel to cancelled
+        cursor.execute(*sql_table.update(
+                [sql_table.state], ['cancelled'],
+                where=sql_table.state == 'cancel'))
+
     @staticmethod
     def default_state():
         return 'draft'
@@ -117,45 +128,75 @@ class Inventory(Workflow, ModelSQL, ModelView):
         return Transaction().context.get('company')
 
     @classmethod
-    def default_lost_found(cls):
-        Location = Pool().get('stock.location')
-        locations = Location.search(cls.lost_found.domain)
-        if len(locations) == 1:
-            return locations[0].id
+    def view_attributes(cls):
+        return super().view_attributes() + [
+            ('/tree', 'visual', If(Eval('state') == 'cancelled', 'muted', '')),
+            ]
 
     @classmethod
     def delete(cls, inventories):
         # Cancel before delete
         cls.cancel(inventories)
         for inventory in inventories:
-            if inventory.state != 'cancel':
-                cls.raise_user_error('delete_cancel', inventory.rec_name)
+            if inventory.state != 'cancelled':
+                raise AccessError(
+                    gettext('stock.msg_inventory_delete_cancel',
+                        inventory=inventory.rec_name))
         super(Inventory, cls).delete(inventories)
 
     @classmethod
     @ModelView.button
     @Workflow.transition('done')
     def confirm(cls, inventories):
-        Move = Pool().get('stock.move')
+        pool = Pool()
+        Move = pool.get('stock.move')
+        Date = pool.get('ir.date')
+        Warning = pool.get('res.user.warning')
+        today_cache = {}
+
+        def in_future(inventory):
+            if inventory.company not in today_cache:
+                with Transaction().set_context(company=inventory.company.id):
+                    today_cache[inventory.company] = Date.today()
+            today = today_cache[inventory.company]
+            if inventory.date > today:
+                return inventory
+        future_inventories = sorted(filter(in_future, inventories))
+        if future_inventories:
+            names = ', '.join(i.rec_name for i in future_inventories[:5])
+            if len(future_inventories) > 5:
+                names + '...'
+            warning_name = (
+                '%s.date_future' % hashlib.md5(
+                    str(future_inventories).encode('utf-8')).hexdigest())
+            if Warning.check(warning_name):
+                raise InventoryFutureWarning(warning_name,
+                    gettext('stock.msg_inventory_date_in_the_future',
+                        inventories=names))
+
         moves = []
         for inventory in inventories:
             keys = set()
             for line in inventory.lines:
                 key = line.unique_key
                 if key in keys:
-                    cls.raise_user_error('unique_line',
-                        (line.rec_name, inventory.rec_name))
+                    raise InventoryValidationError(
+                        gettext('stock.msg_inventory_line_unique',
+                            line=line.rec_name,
+                            inventory=inventory.rec_name))
                 keys.add(key)
                 move = line.get_move()
                 if move:
                     moves.append(move)
         if moves:
             Move.save(moves)
-            Move.do(moves)
+            # Skip MoveFutureWarning as it is newly created moves
+            with Transaction().set_user(0):
+                Move.do(moves)
 
     @classmethod
     @ModelView.button
-    @Workflow.transition('cancel')
+    @Workflow.transition('cancelled')
     def cancel(cls, inventories):
         Line = Pool().get("stock.inventory.line")
         Line.cancel_move([l for i in inventories for l in i.lines])
@@ -163,14 +204,15 @@ class Inventory(Workflow, ModelSQL, ModelView):
     @classmethod
     def create(cls, vlist):
         pool = Pool()
-        Sequence = pool.get('ir.sequence')
         Configuration = pool.get('stock.configuration')
         config = Configuration(1)
         vlist = [x.copy() for x in vlist]
+        default_company = cls.default_company()
         for values in vlist:
             if values.get('number') is None:
-                values['number'] = Sequence.get_id(
-                    config.inventory_sequence.id)
+                values['number'] = config.get_multivalue(
+                    'inventory_sequence',
+                    company=values.get('company', default_company)).get()
         inventories = super(Inventory, cls).create(vlist)
         cls.complete_lines(inventories, fill=False)
         return inventories
@@ -231,14 +273,13 @@ class Inventory(Workflow, ModelSQL, ModelView):
             # Index some data
             product2type = {}
             product2consumable = {}
-            for product in Product.browse([line[1] for line in pbl]):
+            for product in Product.browse({line[1] for line in pbl}):
                 product2type[product.id] = product.type
                 product2consumable[product.id] = product.consumable
 
             # Update existing lines
             for line in inventory.lines:
-                if not (line.product.type == 'goods'
-                        and not line.product.consumable):
+                if line.product.type != 'goods':
                     Line.delete([line])
                     continue
 
@@ -288,9 +329,9 @@ class InventoryLine(ModelSQL, ModelView):
     product = fields.Many2One('product.product', 'Product', required=True,
         domain=[
             ('type', '=', 'goods'),
-            ('consumable', '=', False),
             ], states=_states, depends=_depends)
-    uom = fields.Function(fields.Many2One('product.uom', 'UOM'), 'get_uom')
+    uom = fields.Function(fields.Many2One('product.uom', 'UOM',
+        help="The unit in which the quantity is specified."), 'get_uom')
     unit_digits = fields.Function(fields.Integer('Unit Digits'),
             'get_unit_digits')
     expected_quantity = fields.Float('Expected Quantity', required=True,
@@ -298,41 +339,49 @@ class InventoryLine(ModelSQL, ModelView):
         states={
             'invisible': Eval('id', -1) < 0,
         },
-        depends=['unit_digits'])
+        depends=['unit_digits'],
+        help="The quantity the system calculated should be in the location.")
     quantity = fields.Float('Quantity',
         digits=(16, Eval('unit_digits', 2)),
-        states=_states, depends=['unit_digits'] + _depends)
+        states=_states, depends=['unit_digits'] + _depends,
+        help="The actual quantity found in the location.")
     moves = fields.One2Many('stock.move', 'origin', 'Moves', readonly=True)
     inventory = fields.Many2One('stock.inventory', 'Inventory', required=True,
         ondelete='CASCADE',
         states={
             'readonly': _states['readonly'] & Bool(Eval('inventory')),
             },
-        depends=_depends)
+        depends=_depends,
+        help="The inventory the line belongs to.")
+    inventory_location = fields.Function(
+        fields.Many2One('stock.location', "Location"),
+        'on_change_with_inventory_location',
+        searcher='search_inventory_location')
+    inventory_date = fields.Function(
+        fields.Date("Date"),
+        'on_change_with_inventory_date',
+        searcher='search_inventory_date')
     inventory_state = fields.Function(
-        fields.Selection(INVENTORY_STATES, 'Inventory State',
+        fields.Selection('get_inventory_states', "Inventory State",
         depends=['inventory']),
         'on_change_with_inventory_state')
 
     @classmethod
     def __setup__(cls):
         super(InventoryLine, cls).__setup__()
+        cls.__access__.add('inventory')
         t = cls.__table__()
         cls._sql_constraints += [
             ('check_line_qty_pos', Check(t, t.quantity >= 0),
-                'Line quantity must be positive.'),
+                'stock.msg_inventory_line_quantity_positive'),
             ]
         cls._order.insert(0, ('product', 'ASC'))
-        cls._error_messages.update({
-                'missing_empty_quantity': ('An option for empty quantity is '
-                    'missing for inventory "%(inventory)s".'),
-                'delete_cancel_draft': ('The line "%(line)s" must be on '
-                    'canceled or draft inventory to be deleted.'),
-                })
 
     @classmethod
     def __register__(cls, module_name):
-        cursor = Transaction().connection.cursor()
+        transaction = Transaction()
+        cursor = transaction.connection.cursor()
+        update = transaction.connection.cursor()
         pool = Pool()
         Move = pool.get('stock.move')
         sql_table = cls.__table__()
@@ -346,8 +395,8 @@ class InventoryLine(ModelSQL, ModelView):
         if table.column_exist('move'):
             cursor.execute(*sql_table.select(sql_table.id, sql_table.move,
                     where=sql_table.move != Null))
-            for line_id, move_id in cursor.fetchall():
-                cursor.execute(*move_table.update(
+            for line_id, move_id in cursor:
+                update.execute(*move_table.update(
                         columns=[move_table.origin],
                         values=['%s,%s' % (cls.__name__, line_id)],
                         where=move_table.id == move_id))
@@ -371,6 +420,31 @@ class InventoryLine(ModelSQL, ModelView):
             self.uom = self.product.default_uom
             self.unit_digits = self.product.default_uom.digits
 
+    @fields.depends('inventory', '_parent_inventory.location')
+    def on_change_with_inventory_location(self, name=None):
+        if self.inventory and self.inventory.location:
+            return self.inventory.location.id
+
+    @classmethod
+    def search_inventory_location(cls, name, clause):
+        nested = clause[0].lstrip(name)
+        return [('inventory.' + name + nested,) + tuple(clause[1:])]
+
+    @fields.depends('inventory', '_parent_inventory.date')
+    def on_change_with_inventory_date(self, name=None):
+        if self.inventory:
+            return self.inventory.date
+
+    @classmethod
+    def search_inventory_date(cls, name, clause):
+        return [('inventory.date',) + tuple(clause[1:])]
+
+    @classmethod
+    def get_inventory_states(cls):
+        pool = Pool()
+        Inventory = pool.get('stock.inventory')
+        return Inventory.fields_get(['state'])['state']['selection']
+
     @fields.depends('inventory', '_parent_inventory.state')
     def on_change_with_inventory_state(self, name=None):
         if self.inventory:
@@ -418,9 +492,9 @@ class InventoryLine(ModelSQL, ModelView):
         qty = self.quantity
         if qty is None:
             if self.inventory.empty_quantity is None:
-                self.raise_user_error('missing_empty_quantity', {
-                        'inventory': self.inventory.rec_name,
-                        })
+                raise InventoryValidationError(
+                    gettext('stock.msg_inventory_missing_empty_quantity',
+                        inventory=self.inventory.rec_name))
             if self.inventory.empty_quantity == 'keep':
                 return
             else:
@@ -432,7 +506,12 @@ class InventoryLine(ModelSQL, ModelView):
         if delta_qty == 0.0:
             return
         from_location = self.inventory.location
-        to_location = self.inventory.lost_found
+        to_location = self.inventory.location.lost_found_used
+        if not to_location:
+            raise InventoryValidationError(
+                gettext('stock.msg_inventory_location_missing_lost_found',
+                    inventory=self.inventory.rec_name,
+                    location=self.inventory.location.rec_name))
         if delta_qty < 0:
             (from_location, to_location, delta_qty) = \
                 (to_location, from_location, -delta_qty)
@@ -454,7 +533,7 @@ class InventoryLine(ModelSQL, ModelView):
         '''
         values = {}
         # if nothing changed, no update
-        if self.quantity == self.expected_quantity == quantity:
+        if self.expected_quantity == quantity:
             return values
         values['expected_quantity'] = quantity
         return values
@@ -472,10 +551,11 @@ class InventoryLine(ModelSQL, ModelView):
     @classmethod
     def delete(cls, lines):
         for line in lines:
-            if line.inventory_state not in {'cancel', 'draft'}:
-                cls.raise_user_error('delete_cancel_draft', {
-                        'line': line.rec_name,
-                        })
+            if line.inventory_state not in {'cancelled', 'draft'}:
+                raise AccessError(
+                    gettext('stock.msg_inventory_line_delete_cancel',
+                        line=line.rec_name,
+                        inventory=line.inventory.rec_name))
         super(InventoryLine, cls).delete(lines)
 
 
@@ -498,28 +578,21 @@ class Count(Wizard):
             ])
     add = StateTransition()
 
-    @classmethod
-    def __setup__(cls):
-        super(Count, cls).__setup__()
-        cls._error_messages.update({
-                'create_line': "No existing line found for %(search)s.",
-                })
-
     def default_quantity(self, fields):
         pool = Pool()
-        Inventory = pool.get('stock.inventory')
         InventoryLine = pool.get('stock.inventory.line')
-        context = Transaction().context
-        inventory = Inventory(context['active_id'])
+        Warning = pool.get('res.user.warning')
         values = {}
-        lines = InventoryLine.search(self.get_line_domain(inventory), limit=1)
+        lines = InventoryLine.search(
+            self.get_line_domain(self.record), limit=1)
         if not lines:
             warning_name = '%s.%s.count_create' % (
-                inventory, self.search.search)
-            self.raise_user_warning(warning_name, 'create_line', {
-                    'search': self.search.search.rec_name,
-                    })
-            line, = InventoryLine.create([self.get_line_values(inventory)])
+                self.record, self.search.search)
+            if Warning.check(warning_name):
+                raise InventoryCountWarning(warning_name,
+                    gettext('stock.msg_inventory_count_create_line',
+                        search=self.search.search.rec_name))
+            line, = InventoryLine.create([self.get_line_values(self.record)])
         else:
             line, = lines
         values['line'] = line.id
@@ -527,7 +600,7 @@ class Count(Wizard):
         values['uom'] = line.uom.id
         values['unit_digits'] = line.unit_digits
         if line.uom.rounding == 1:
-            values['quantity_added'] = 1
+            values['quantity'] = 1
         return values
 
     def get_line_domain(self, inventory):
@@ -550,12 +623,12 @@ class Count(Wizard):
         return values
 
     def transition_add(self):
-        if self.quantity.line and self.quantity.quantity_added:
+        if self.quantity.line and self.quantity.quantity:
             line = self.quantity.line
             if line.quantity:
-                line.quantity += self.quantity.quantity_added
+                line.quantity += self.quantity.quantity
             else:
-                line.quantity = self.quantity.quantity_added
+                line.quantity = self.quantity.quantity
             line.save()
         return 'search'
 
@@ -575,7 +648,8 @@ class CountSearch(ModelView):
                     ('consumable', '=', False),
                     ],
                 [])],
-        depends=['search_model'])
+        depends=['search_model'],
+        help="The item that's counted.")
     search_model = fields.Function(fields.Selection(
         'get_search_models', "Search Model"),
         'on_change_with_search_model')
@@ -601,19 +675,22 @@ class CountQuantity(ModelView):
     line = fields.Many2One(
         'stock.inventory.line', "Line", readonly=True, required=True)
     product = fields.Many2One('product.product', "Product", readonly=True)
-    uom = fields.Many2One('product.uom', "UOM", readonly=True)
-    quantity_resulting = fields.Float(
-        "Resulting Quantity", digits=(16, Eval('unit_digits', 2)),
-        readonly=True, depends=['unit_digits'])
+    uom = fields.Many2One('product.uom', "UOM", readonly=True,
+        help="The unit in which the quantities are specified.")
+    total_quantity = fields.Float(
+        "Total Quantity", digits=(16, Eval('unit_digits', 2)),
+        readonly=True, depends=['unit_digits'],
+        help="The total amount of the line counted so far.")
 
-    quantity_added = fields.Float(
-        "Added Quantity", digits=(16, Eval('unit_digits', 2)), required=True,
-        depends=['unit_digits'])
+    quantity = fields.Float(
+        "Quantity", digits=(16, Eval('unit_digits', 2)), required=True,
+        depends=['unit_digits'],
+        help="The quantity to add to the existing count.")
 
     unit_digits = fields.Integer("Unit Digits", readonly=True)
 
-    @fields.depends('quantity_added', 'line')
-    def on_change_quantity_added(self):
+    @fields.depends('quantity', 'line')
+    def on_change_quantity(self):
         if self.line:
-            self.quantity_resulting = (
-                (self.line.quantity or 0) + (self.quantity_added or 0))
+            self.total_quantity = (
+                (self.line.quantity or 0) + (self.quantity or 0))
